/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 * 
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * This file contains the template code meant to be used in the tutorial 
 * "A beginner's tutorial: Advertising" found on http://devzone.nordicsemi.com/tutorials/
 *
 */

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "nrf51_bitfields.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_conn_params.h"
#include "app_scheduler.h"
#include "softdevice_handler.h"
#include "app_timer_appsh.h"
#include "nrf_drv_gpiote.h"
#include "pca10028.h"
#include "nrf_ecb.h"
#include "nrf_delay.h"
#include "pstorage.h"


#define IS_SRVC_CHANGED_CHARACT_PRESENT 0                                           /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/
#define APP_ADV_INTERVAL                0x0200 //160                                /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      2                                           /**< The advertising timeout (in units of seconds). */

#define APP_TIMER_PRESCALER             0                                           /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_MAX_TIMERS            2                                           /**< Maximum number of simultaneously created timers. */
#define APP_TIMER_OP_QUEUE_SIZE         4                                           /**< Size of timer operation queues. */

#define APP_BEACON_INFO_LENGTH          0x11                              /**< Total length of information advertised by the Beacon. */
#define APP_ADV_DATA_LENGTH             0x15                              /**< Length of manufacturer specific data in the advertisement. */
#define APP_DEVICE_TYPE                 0x02                              /**< 0x02 refers to Beacon. */
#define APP_MEASURED_RSSI               0xC3                              /**< The Beacon's measured RSSI at 1 meter distance in dBm. */
#define APP_COMPANY_IDENTIFIER          0x0059                            /**< Company identifier for Nordic Semiconductor ASA. as per www.bluetooth.org. */
#define APP_MAJOR_VALUE                 0x01, 0x02                        /**< Major value used to identify Beacons. */ 
#define APP_MINOR_VALUE                 0x03, 0x04                        /**< Minor value used to identify Beacons. */ 

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(500, UNIT_1_25_MS)            /**< Minimum acceptable connection interval (0.5 seconds). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(1000, UNIT_1_25_MS)           /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds). */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000, APP_TIMER_PRESCALER)  /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000, APP_TIMER_PRESCALER) /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */

#define APP_GPIOTE_MAX_USERS            1                                           /**< Maximum number of users of the GPIOTE handler. */

#define BUTTON_DETECTION_DELAY          APP_TIMER_TICKS(50, APP_TIMER_PRESCALER)    /**< Delay from a GPIOTE event until a button is reported as pushed (in number of timer ticks). */
#define DEVICE_NAME                     "Active_BLE"  
#define SEC_PARAM_BOND                  1                                           /**< Perform bonding. */
#define SEC_PARAM_MITM                  0                                           /**< Man In The Middle protection not required. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_NONE                        /**< No I/O capabilities. */
#define SEC_PARAM_OOB                   0                                           /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                           /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                          /**< Maximum encryption key size. */

#define SCAN_REQUEST_NOTIFICATION				1
#define TIMER_TICKS_PER_SECOND  				32768
#define RTC_TICKS_BITS									16

typedef struct
{
    int32_t seconds;       // seconds (signed for delta) 
    int32_t ticks;         // in TIMESTAMP_TICKS_PER_SECOND 
} timestamp_t;

static timestamp_t 								  		timestamp;
static app_timer_id_t 						      local_timer;
static bool loop_open									  = false;
static bool loop_close								  = false;
static ble_gap_sec_params_t             m_sec_params;                               /**< Security requirements for this application. */
static uint16_t                         m_conn_handle = BLE_CONN_HANDLE_INVALID;    /**< Handle of the current connection. */
static ble_gap_adv_params_t 						adv_params;
static ble_advdata_manuf_data_t 				manuf_specific_data;
static uint8_t form_packet[16]					={0}; //packet being formed
static uint8_t packet_form[16]					={0}; //packet after encrption
static nrf_ecb_hal_data_t								nrf_ecb_hal_data;
static ble_uuid_t          adv_uuids[] = {{0x1234, BLE_UUID_TYPE_BLE}};    // Random example User Unique Identifier
static ble_gap_addr_t ble_base_station;
static uint8_t ble_addr[] = {0x57, 0x1C, 0x73, 0x4A, 0xEA, 0x40};
static uint8_t pstorage_wait_flag = 0;
static pstorage_block_t pstorage_wait_handle = 0;


#define SCHED_MAX_EVENT_DATA_SIZE       MAX(APP_TIMER_SCHED_EVT_SIZE, sizeof(nrf_drv_gpiote_pin_t))                  /**< Maximum size of scheduler events. Note that scheduler BLE stack events do not contain any data, as the events are being pulled from the stack in the event handler. */
#define SCHED_QUEUE_SIZE                10                                          /**< Maximum number of events in the scheduler queue. */

static uint8_t m_beacon_info[APP_BEACON_INFO_LENGTH] =                    /**< Information advertised by the Beacon. */
{
    APP_BEACON_INFO_LENGTH, // Manufacturer specific information. Specifies the length of the                
};

static void example_cb_handler(pstorage_handle_t  * handle,
															 uint8_t              op_code,
                               uint32_t             result,
                               uint8_t            * p_data,
                               uint32_t             data_len)
{
		if(handle->block_id == pstorage_wait_handle) { pstorage_wait_flag = 0; }  //If we are waiting for this callback, clear the wait flag.
	
		switch(op_code)
		{
			case PSTORAGE_LOAD_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
						
				 }
				 else
				 {
						
				 }
				 break;
			case PSTORAGE_STORE_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
						
				 }
				 else
				 {
						
				 }
				 break;				 
			case PSTORAGE_UPDATE_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
					
				 }
				 else
				 {
					
				 }
				 break;
			case PSTORAGE_CLEAR_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
					
				 }
				 else
				 {
					
				 }
				 break;
			default:
				 break;				 
		}			
}

static void pstorage_test_store_and_update(void)
{
		pstorage_handle_t       handle;
		pstorage_handle_t				block_0_handle;
		pstorage_handle_t				block_1_handle;
		pstorage_handle_t				block_2_handle;
		pstorage_module_param_t param;
		uint8_t                 source_data_0[16] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F};
		uint8_t                 source_data_1[16] = {0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F};
		uint8_t                 source_data_2[16] = {0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F};
		uint8_t                 source_data_9[16] = {0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F};
		uint8_t                 dest_data_0[16];
		uint8_t                 dest_data_1[16];
		uint8_t                 dest_data_2[16];
		uint32_t                retval;
//		
//		retval = pstorage_init();
//		if(retval != NRF_SUCCESS)
//		{
//				bsp_indication_set(BSP_INDICATE_FATAL_ERROR);
//		}
//	     
		param.block_size  = 16;                   //Select block size of 16 bytes
		param.block_count = 10;                   //Select 10 blocks, total of 160 bytes
		param.cb          = example_cb_handler;   //Set the pstorage callback handler
			
		retval = pstorage_register(&param, &handle);
		APP_ERROR_CHECK(retval);
		
		//Get block identifiers
		pstorage_block_identifier_get(&handle, 0, &block_0_handle);
		pstorage_block_identifier_get(&handle, 1, &block_1_handle);
		pstorage_block_identifier_get(&handle, 2, &block_2_handle);
		
				//Clear 100 bytes, i.e. one block
		retval = pstorage_clear(&block_0_handle, 16);                       
		APP_ERROR_CHECK(retval);
		
		//Store data to three blocks. Wait for the last store operation to finish before reading out the data.
		pstorage_store(&block_0_handle, source_data_0, 16, 0);     //Write to flash, only one block is allowed for each pstorage_store command
		APP_ERROR_CHECK(retval);
		
		pstorage_store(&block_1_handle, source_data_1, 16, 0);     //Write to flash, only one block is allowed for each pstorage_store command		
		pstorage_wait_handle = block_2_handle.block_id;            //Specify which pstorage handle to wait for
		pstorage_wait_flag = 1;                                    //Set the wait flag. Cleared in the example_cb_handler
		pstorage_store(&block_2_handle, source_data_2, 16, 0);     //Write to flash
		while(pstorage_wait_flag) { power_manage(); }              //Sleep until store operation is finished.
		
		pstorage_load(dest_data_0, &block_0_handle, 16, 0);				 //Read from flash, only one block is allowed for each pstorage_load command
		pstorage_load(dest_data_1, &block_1_handle, 16, 0);				 //Read from flash
		pstorage_load(dest_data_2, &block_2_handle, 16, 0);			   //Read from flash
		
		pstorage_wait_handle = block_0_handle.block_id;            //Specify which pstorage handle to wait for
		pstorage_wait_flag = 1;                                    //Set the wait flag. Cleared in the example_cb_handler
		pstorage_clear(&block_0_handle, 32);                       //Clear 32 bytes
		while(pstorage_wait_flag) { power_manage(); }              //Sleep until store operation is finished.
		
		pstorage_load(dest_data_0, &block_0_handle, 16, 0);				 //Read from flash, only one block is allowed for each pstorage_load command
		pstorage_load(dest_data_1, &block_1_handle, 16, 0);				 //Read from flash
		pstorage_load(dest_data_2, &block_2_handle, 16, 0);			   //Read from flash
		
		pstorage_wait_handle = block_0_handle.block_id;            //Specify which pstorage handle to wait for 
		pstorage_wait_flag = 1;                                    //Set the wait flag. Cleared in the example_cb_handler
		pstorage_update(&block_0_handle, source_data_9, 16, 0);    //update flash block 0
		while(pstorage_wait_flag) { power_manage(); }              //Sleep until update operation is finished.
		
		pstorage_load(dest_data_0, &block_0_handle, 16, 0);				 //Read from flash, only one block is allowed for each pstorage_load command
		pstorage_load(dest_data_1, &block_1_handle, 16, 0);				 //Read from flash
		pstorage_load(dest_data_2, &block_2_handle, 16, 0);			   //Read from flash
}



/**@brief Function for the Timer initialization.
 * This function sets up software timers. The clock source for the timers
is the nRF51s Real Time Clock 1 (RTC1). In this example we initialize 2 timers 
(see the #define APP_TIMER_MAX_TIMERS). These are used to time the bluetooth connection intervals
 */
static void timers_init(void)
{
    // Initialize timer module, making it use the scheduler
    APP_TIMER_APPSH_INIT(APP_TIMER_PRESCALER, APP_TIMER_MAX_TIMERS, APP_TIMER_OP_QUEUE_SIZE, true);
}

static void ppi_init(void)
{
    // When using the Timeslot advertiser-scanner library the only PPI channels available are 3 and 7
    
    // Configure PPI channel 0 to trigger COUNT on TIMER1 when COMPARE[1] matches on RTC1
	NRF_PPI->CH[7].EEP = (uint32_t)(&NRF_RTC1->EVENTS_COMPARE[1]);
	NRF_PPI->CH[7].TEP = (uint32_t)(&NRF_TIMER1->TASKS_COUNT);
	
	// Configure PPI channel 1 to trigger CLEAR on RTC1 when COMPARE[1] matches on RTC1
	NRF_PPI->CH[3].EEP = (uint32_t)(&NRF_RTC1->EVENTS_COMPARE[1]);
	NRF_PPI->CH[3].TEP = (uint32_t)(&NRF_RTC1->TASKS_CLEAR);
	
	// Enable only PPI channels 0 and 1.
	NRF_PPI->CHEN = (PPI_CHEN_CH3_Enabled << PPI_CHEN_CH3_Pos) | (PPI_CHEN_CH7_Enabled << PPI_CHEN_CH7_Pos);
}

static void timer1_init(void)
{
	// Configure TIMER1 in COUNT mode and 32 bit
	NRF_TIMER1->MODE = TIMER_MODE_MODE_Counter;
	NRF_TIMER1->BITMODE = TIMER_BITMODE_BITMODE_32Bit;  
	NRF_TIMER1->TASKS_CLEAR = 1;
	NRF_TIMER1->TASKS_START = 1;	
}

void timestampGet(timestamp_t *ts)
{
    uint32_t overflow, ticks, s, t;

    app_timer_ticks(&overflow, &ticks);

    s = ticks / TIMER_TICKS_PER_SECOND; // becomes >> 
    t = ticks % TIMER_TICKS_PER_SECOND; // becomes & 

    ts->seconds =
        overflow * ((1 << RTC_TICKS_BITS) / TIMER_TICKS_PER_SECOND) + s;
    ts->ticks = t * (APP_TIMER_PRESCALER + 1); // timestamp_t.ticks always mod 32768 
}

void aes_encrypt()
{
	 memset (&nrf_ecb_hal_data, 0, sizeof(nrf_ecb_hal_data));
	
	 for(int i=0;i<16;i++)
	 nrf_ecb_hal_data.cleartext[i] = form_packet[i];
	
	 for(int i=0;i<2;i++)
	 nrf_ecb_hal_data.key[i] = 'b';
	
	 for(int i=3;i<16;i++)
	 nrf_ecb_hal_data.key[i] = 0x00;
	
	 uint32_t	err_code = sd_ecb_block_encrypt(& nrf_ecb_hal_data);
	 APP_ERROR_CHECK(err_code);
	
	 for(int i=0;i<16;i++)
	 packet_form[i] = nrf_ecb_hal_data.ciphertext[i];
	
}

/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void)
{
    uint32_t                err_code;
	
	// Declearing parameter structs. Try to go to the struct definitions to get
	// more information about what parameters they contain
    ble_gap_conn_params_t   gap_conn_params; 	// Struct to store GAP connection parameters like max min connection interval etc
    ble_gap_conn_sec_mode_t sec_mode;			// Struct to store security parameters 

	// A simple macro that sets the Security Mode and Level bits in sec_mode
	// to require no protection (open link)
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

	// Store the device name and security mode in the SoftDevice. Our name is defined to "HelloWorld" in the beginning of this file
    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *)DEVICE_NAME,
                                          strlen(DEVICE_NAME));
		APP_ERROR_CHECK(err_code);// Check for errors

	// Always initialize all fields in structs to zero or you might get unexpected behaviour
    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

	// Populate the GAP connection parameter struct
    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

	// Set GAP Peripheral Preferred Connection Parameters
    // The device use these prefered values when negotiating connection terms with another device
    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);// Check for errors
											
	// Set appearence										  
		sd_ble_gap_appearance_set(0);
		APP_ERROR_CHECK(err_code);// Check for errors
}

/**@brief Function for initializing the Advertising functionality.
 *
 * @details Encodes the required advertising data and passes it to the stack.
 *          Also builds a structure to be passed to the stack when starting advertising.
 */
static void advertising_init(void)
{
    uint32_t            err_code;
    ble_advdata_t       advdata;    // Struct containing advertising parameters
    uint8_t             flags       = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE; // Defines how your device is seen (or not seen) by other devices
   
		form_packet[0] =  'L';
	  form_packet[1] =  '1';
	
		aes_encrypt();
	  
		for(int i=0;i<16;i++)
		m_beacon_info[i+1] = packet_form[i];
	
		manuf_specific_data.data.p_data = (uint8_t *) m_beacon_info;
    manuf_specific_data.data.size   = APP_BEACON_INFO_LENGTH;
				
    // Populate the advertisement packet
    // Always initialize all fields in structs to zero or you might get unexpected behaviour
    memset(&advdata, 0, sizeof(advdata));
    
    advdata.name_type               = BLE_ADVDATA_NO_NAME;                                             
    advdata.flags                   = flags;                                     // Must be included, but not discussed in this tutorial
		advdata.p_manuf_specific_data   = &manuf_specific_data;
		advdata.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]); // Must be included, but not discussed in this tutorial
		advdata.uuids_complete.p_uuids  = adv_uuids;                                // Must be included, but not discussed in this tutorial
	
    // Set the advertisement and scan response packet. 
    err_code = ble_advdata_set(&advdata, NULL);
    APP_ERROR_CHECK(err_code);// Check for errors
}


/**@brief Function for starting advertising.
 */
static void advertising_start(void)
{
    uint32_t             err_code;
		
    // Always initialize all fields in structs to zero or you might get unexpected behaviour
    memset(&adv_params, 0, sizeof(adv_params));

    adv_params.type        = BLE_GAP_ADV_TYPE_ADV_IND;
    adv_params.p_peer_addr = NULL;
    adv_params.fp          = BLE_GAP_ADV_FP_ANY;
    adv_params.interval    = APP_ADV_INTERVAL;
    adv_params.timeout     = APP_ADV_TIMEOUT_IN_SECONDS;

    err_code = sd_ble_gap_adv_start(&adv_params);
    APP_ERROR_CHECK(err_code);// Check for errors
}


void gpiote_handler_init() 
{
		if(!nrf_drv_gpiote_is_init())
		{ 
			nrf_drv_gpiote_init();
		} 
}

void button_handler(nrf_drv_gpiote_pin_t pin)
{
		advertising_init();
		advertising_start();
}

void button_scheduler_event_handler(void *p_event_data,uint16_t event_size)
{
		button_handler(*((nrf_drv_gpiote_pin_t*)p_event_data));
}	

void in_pin_handler(nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action)
{
	switch(pin) 
	{
		case OPEN_BIN:
				 if(!(loop_open))
					{
						loop_open = true;
						form_packet[2] = 0x00;
						timestampGet(&timestamp);
						form_packet[3] = (uint8_t)(timestamp.seconds>>24);
						form_packet[4] = (uint8_t)(timestamp.seconds>>16);
						form_packet[5] = (uint8_t)(timestamp.seconds>>8);
						form_packet[6] = (uint8_t)(timestamp.seconds);						
						app_sched_event_put(&pin,sizeof(pin),button_scheduler_event_handler);
					}
				 break;

		case CLOSE_BIN:
					if(!(loop_close))
					{
						loop_close = true;
						form_packet[2] = 0xFF;
						timestampGet(&timestamp);
						form_packet[3] = (uint8_t)(timestamp.seconds>>24);
						form_packet[4] = (uint8_t)(timestamp.seconds>>16);
						form_packet[5] = (uint8_t)(timestamp.seconds>>8);
						form_packet[6] = (uint8_t)(timestamp.seconds); 
						app_sched_event_put(&pin,sizeof(pin),button_scheduler_event_handler);
					}
				 break;
		
		default:
			;
	}
}
	

void gpio_init(void)
{
		uint32_t error_code;
		
		error_code = nrf_drv_gpiote_init();
		APP_ERROR_CHECK(error_code);
		
		nrf_drv_gpiote_in_config_t in_config = GPIOTE_CONFIG_IN_SENSE_HITOLO(true);
		in_config.pull = NRF_GPIO_PIN_PULLUP;
		
		error_code = nrf_drv_gpiote_in_init(OPEN_BIN,&in_config,in_pin_handler);
		APP_ERROR_CHECK(error_code);
		
		error_code = nrf_drv_gpiote_in_init(CLOSE_BIN,&in_config,in_pin_handler);
		APP_ERROR_CHECK(error_code);
			
		nrf_drv_gpiote_in_event_enable(OPEN_BIN,true);
		
		nrf_drv_gpiote_in_event_enable(CLOSE_BIN,true);
}


/**@brief Function for initializing security parameters.
 */
static void sec_params_init(void)
{
    m_sec_params.bond         = SEC_PARAM_BOND;
    m_sec_params.mitm         = SEC_PARAM_MITM;
    m_sec_params.io_caps      = SEC_PARAM_IO_CAPABILITIES;
    m_sec_params.oob          = SEC_PARAM_OOB;
    m_sec_params.min_key_size = SEC_PARAM_MIN_KEY_SIZE;
    m_sec_params.max_key_size = SEC_PARAM_MAX_KEY_SIZE;
}


/**@brief Function for handling the Connection Parameters Module.
 *
 * @details This function will be called for all events in the Connection Parameters Module which
 *          are passed to the application.
 *          @note All this function does is to disconnect. This could have been done by simply
 *                setting the disconnect_on_fail config parameter, but instead we use the event
 *                handler mechanism to demonstrate its use.
 *
 * @param[in]   p_evt   Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
    uint32_t err_code; // Variable used to check for any errors returned by functions

    if(p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);// Check for errors
    }
}


/**@brief Function for handling a Connection Parameters error.
 *
 * @param[in]   nrf_error   Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}


/**@brief Function for initializing the Connection Parameters module.
 */
static void conn_params_init(void)
{
    uint32_t               err_code;
    ble_conn_params_init_t cp_init;
	// Always initialize all fields in structs to zero or you might get unexpected behaviour
    memset(&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = NULL;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = false;
    cp_init.evt_handler                    = on_conn_params_evt;
    cp_init.error_handler                  = conn_params_error_handler;

    err_code = ble_conn_params_init(&cp_init);
    // Check for errors
	APP_ERROR_CHECK(err_code);// Check for errors
}

static void sleep_mode_enter(void)
{
    uint32_t err_code;
	
		loop_open 	= false;
	  loop_close  = false;
		err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}

static void local_timer_handler(void * p_context)
{

}

void timer_start()
{
		uint32_t error_code;
		
		error_code = app_timer_create(&local_timer,APP_TIMER_MODE_REPEATED,local_timer_handler);
		APP_ERROR_CHECK(error_code);
		
		error_code = app_timer_start(local_timer,5,NULL);//APP_TIMER_TICKS(1000, APP_TIMER_PRESCALER)
		APP_ERROR_CHECK(error_code);
}

static bool BASE_STATION_PRESENT = false;

static void stop_advertising(void)
{
		uint32_t err_code;
	
		BASE_STATION_PRESENT++;
	
		err_code = sd_ble_gap_adv_stop();
		APP_ERROR_CHECK(err_code);
	
		err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t                         err_code;
    static ble_gap_evt_auth_status_t m_auth_status;
    bool                             master_id_matches;
    ble_gap_sec_kdist_t *            p_distributed_keys;
    ble_gap_enc_info_t *             p_enc_info;
    ble_gap_irk_t *                  p_id_info;
    ble_gap_sign_info_t *            p_sign_info;

    static ble_gap_enc_key_t         m_enc_key;           /**< Encryption Key (Encryption Info and Master ID). */
    static ble_gap_id_key_t          m_id_key;            /**< Identity Key (IRK and address). */
    static ble_gap_sign_info_t       m_sign_key;          /**< Signing Key (Connection Signature Resolving Key). */
    static ble_gap_sec_keyset_t      m_keys = {.keys_periph = {&m_enc_key, &m_id_key, &m_sign_key}};

    switch (p_ble_evt->header.evt_id)
    {
			
        case BLE_GAP_EVT_CONNECTED:
					   err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
							APP_ERROR_CHECK(err_code);// Check for errors
						m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            m_conn_handle = BLE_CONN_HANDLE_INVALID;         
            advertising_start();
            break;
				
        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            err_code = sd_ble_gap_sec_params_reply(m_conn_handle,
                                                   BLE_GAP_SEC_STATUS_SUCCESS,
                                                   &m_sec_params,
                                                   &m_keys);
            APP_ERROR_CHECK(err_code);// Check for errors
            break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            err_code = sd_ble_gatts_sys_attr_set(m_conn_handle,
                                                 NULL,
                                                 0,
                                                 BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
            APP_ERROR_CHECK(err_code);// Check for errors
            break;

        case BLE_GAP_EVT_AUTH_STATUS:
            m_auth_status = p_ble_evt->evt.gap_evt.params.auth_status;
            break;

        case BLE_GAP_EVT_SEC_INFO_REQUEST:
            master_id_matches  = memcmp(&p_ble_evt->evt.gap_evt.params.sec_info_request.master_id,
                                        &m_enc_key.master_id,
                                        sizeof(ble_gap_master_id_t)) == 0;
            p_distributed_keys = &m_auth_status.kdist_periph;

            p_enc_info  = (p_distributed_keys->enc  && master_id_matches) ? &m_enc_key.enc_info : NULL;
            p_id_info   = (p_distributed_keys->id   && master_id_matches) ? &m_id_key.id_info   : NULL;
            p_sign_info = (p_distributed_keys->sign && master_id_matches) ? &m_sign_key         : NULL;

            err_code = sd_ble_gap_sec_info_reply(m_conn_handle, p_enc_info, p_id_info, p_sign_info);
            APP_ERROR_CHECK(err_code);// Check for errors
            break;

				case BLE_GAP_EVT_SCAN_REQ_REPORT:
					
			if(p_ble_evt->evt.gap_evt.params.scan_req_report.peer_addr.addr[0] == ble_base_station.addr[0]&& 
				 p_ble_evt->evt.gap_evt.params.scan_req_report.peer_addr.addr[1] == ble_base_station.addr[1]&&
				 p_ble_evt->evt.gap_evt.params.scan_req_report.peer_addr.addr[2] == ble_base_station.addr[2]&&
				 p_ble_evt->evt.gap_evt.params.scan_req_report.peer_addr.addr[3] == ble_base_station.addr[3]&&
				 p_ble_evt->evt.gap_evt.params.scan_req_report.peer_addr.addr[4] == ble_base_station.addr[4]&&
				 p_ble_evt->evt.gap_evt.params.scan_req_report.peer_addr.addr[5] == ble_base_station.addr[5])
						BASE_STATION_PRESENT = true;
//				stop_advertising();
//						nrf_delay_ms(100);
//						sleep_mode_enter();
                // Go to system-off mode (this function will not return; wakeup will cause a reset)                
                //err_code = sd_power_system_off();
                //APP_ERROR_CHECK(err_code);// Check for errors
          
						break;
				
        case BLE_GAP_EVT_TIMEOUT:					
            if (p_ble_evt->evt.gap_evt.params.timeout.src == BLE_GAP_TIMEOUT_SRC_ADVERTISING)
            {
							
							sleep_mode_enter();
                // Go to system-off mode (this function will not return; wakeup will cause a reset)                
                //err_code = sd_power_system_off();
                //APP_ERROR_CHECK(err_code);// Check for errors
            }
            break;

        default:
            // No implementation needed.
            break;
    }
}

/**@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    on_ble_evt(p_ble_evt);
    ble_conn_params_on_ble_evt(p_ble_evt);
}

static void sys_evt_dispatch(uint32_t sys_evt)
{
		pstorage_sys_event_handler(sys_evt);
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    uint32_t err_code; // Variable used to check for any errors returned by functions

    // Initialize the SoftDevice handler module using the low frequency crystal oscillator with 20 ppm accuracy
    SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_RC_250_PPM_TEMP_1000MS_CALIBRATION, NULL);

    // Enable BLE stack 
	// Declaration of a simple struct containing ble parameters
    ble_enable_params_t ble_enable_params;
	
	// Always initialize all fields in structs to zero or you might get unexpected behaviour
    memset(&ble_enable_params, 0, sizeof(ble_enable_params));
	
	// This parameter defines whether or not our device might reconfigure its services during its lifetime. 
	// In our case we set the value to 0 meaning that no service or attributes will ever change.
    ble_enable_params.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
	
	// This call initializes the bluetooth stack, no other BLE related call can be called before this one has been executed.
    err_code = sd_ble_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);// Check for errors
    
    // Tell the softdevice which function is to be called i case of a BLE event. In this case: ble_evt_dispatch()
    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);// Check for errors

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);

}


/**@brief Function for the Event Scheduler initialization.
 */
static void scheduler_init(void)
{
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);
}

/**@brief Function for the Power manager.
 */
static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);// Check for errors
}

/**@brief Function for application main entry.
 */
int main(void)
{
	 uint32_t err_code;
	
	 NRF_POWER->RAMON = POWER_RAMON_ONRAM0_RAM0On   << POWER_RAMON_ONRAM0_Pos
                   | POWER_RAMON_ONRAM1_RAM1On   << POWER_RAMON_ONRAM1_Pos
                   | POWER_RAMON_OFFRAM0_RAM0Off << POWER_RAMON_OFFRAM0_Pos
                   | POWER_RAMON_OFFRAM1_RAM1Off << POWER_RAMON_OFFRAM1_Pos;
	

    // Initialize
		gpio_init();
		timer1_init();
	  ppi_init();
		timers_init();
    ble_stack_init();
		pstorage_init();
	  scheduler_init();
		gap_params_init();
    advertising_init();
    conn_params_init();
    sec_params_init();
	
	for(int i=0;i<6;i++)
	ble_base_station.addr[i] = ble_addr[i]; 
	
#ifdef SCAN_REQUEST_NOTIFICATION
		const ble_opt_t scan_req_opt={ .gap_opt = { .scan_req_report = { .enable=1} } };
		err_code = sd_ble_opt_set(BLE_GAP_OPT_SCAN_REQ_REPORT, &scan_req_opt);
		APP_ERROR_CHECK(err_code);
#endif

    // Start execution
		timer_start();
    advertising_start();

    // Enter main loop
    for (;;)
    {
        app_sched_execute();
        power_manage();
    }
}

/**
 * @}
 */
